# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import functools
import threading
import time
import traceback
from abc import ABC
from urllib.parse import urlparse, urljoin

from jose import ExpiredSignatureError
from keycloak.exceptions import KeycloakClientError
from mako.exceptions import RichTraceback
from tornado import locale
from tornado.ioloop import IOLoop
from tornado.web import RequestHandler

import sickrage
from sickrage.core import helpers
from sickrage.core.helpers import is_ip_whitelisted


class BaseHandler(RequestHandler, ABC):
    def __init__(self, application, request, **kwargs):
        super(BaseHandler, self).__init__(application, request, **kwargs)
        threading.currentThread().setName('TORNADO')

        self.startTime = time.time()

    def get_user_locale(self):
        return locale.get(sickrage.app.config.gui_lang)

    def write_error(self, status_code, **kwargs):
        if self.settings.get("debug") and "exc_info" in kwargs:
            exc_info = kwargs["exc_info"]
            trace_info = ''.join(["%s<br>" % line for line in traceback.format_exception(*exc_info)])
            request_info = ''.join(["<strong>%s</strong>: %s<br>" % (k, self.request.__dict__[k]) for k in self.request.__dict__.keys()])
            error = exc_info[1]

            sickrage.app.log.error(error)

            self.set_header('Content-Type', 'text/html')
            return self.write("""<html>
                             <title>{error}</title>
                             <body>
                                <button onclick="window.location='{webroot}/logs/';">View Log(Errors)</button>
                                <button onclick="window.location='{webroot}/home/restart?pid={pid}&force=1';">Restart SiCKRAGE</button>
                                <button onclick="window.location='{webroot}/logout';">Logout</button>
                                <h2>Error</h2>
                                <p>{error}</p>
                                <h2>Traceback</h2>
                                <p>{traceback}</p>
                                <h2>Request Info</h2>
                                <p>{request}</p>
                             </body>
                           </html>""".format(pid=sickrage.app.pid,
                                             error=error,
                                             traceback=trace_info,
                                             request=request_info,
                                             webroot=sickrage.app.config.web_root))

    def get_current_user(self):
        if is_ip_whitelisted(self.request.remote_ip):
            return True
        elif sickrage.app.config.sso_auth_enabled and sickrage.app.auth_server.health:
            try:
                access_token = self.get_secure_cookie('_sr_access_token')
                refresh_token = self.get_secure_cookie('_sr_refresh_token')
                if not all([access_token, refresh_token]):
                    return

                certs = sickrage.app.auth_server.certs()
                if not certs:
                    return

                try:
                    return sickrage.app.auth_server.decode_token(access_token.decode("utf-8"), certs)
                except (KeycloakClientError, ExpiredSignatureError):
                    token = sickrage.app.auth_server.refresh_token(refresh_token.decode("utf-8"))
                    if not token:
                        return

                    self.set_secure_cookie('_sr_access_token', token['access_token'])
                    self.set_secure_cookie('_sr_refresh_token', token['refresh_token'])
                    return sickrage.app.auth_server.decode_token(token['access_token'], certs)
            except Exception as e:
                return
        elif sickrage.app.config.local_auth_enabled:
            cookie = self.get_secure_cookie('_sr').decode() if self.get_secure_cookie('_sr') else None
            if cookie == sickrage.app.config.api_key:
                return True

    def render_string(self, template_name, **kwargs):
        template_kwargs = {
            'title': "",
            'header': "",
            'topmenu': "",
            'submenu': "",
            'controller': "home",
            'action': "index",
            'srPID': sickrage.app.pid,
            'srHttpsEnabled': sickrage.app.config.enable_https or bool(self.request.headers.get('X-Forwarded-Proto') == 'https'),
            'srHost': self.request.headers.get('X-Forwarded-Host', self.request.host.split(':')[0]),
            'srHttpPort': self.request.headers.get('X-Forwarded-Port', sickrage.app.config.web_port),
            'srHttpsPort': sickrage.app.config.web_port,
            'srHandleReverseProxy': sickrage.app.config.handle_reverse_proxy,
            'srThemeName': sickrage.app.config.theme_name,
            'srDefaultPage': sickrage.app.config.default_page,
            'srWebRoot': sickrage.app.config.web_root,
            'srLocale': self.get_user_locale().code,
            'srLocaleDir': sickrage.LOCALE_DIR,
            'srStartTime': self.startTime,
            'makoStartTime': time.time(),
            'overall_stats': None,
            'torrent_webui_url': helpers.torrent_webui_url(),
            'application': self.application,
            'request': self.request,
        }

        template_kwargs.update(self.get_template_namespace())
        template_kwargs.update(kwargs)

        try:
            return self.application.settings['templates'][template_name].render_unicode(**template_kwargs)
        except Exception:
            kwargs['title'] = _('HTTP Error 500')
            kwargs['header'] = _('HTTP Error 500')
            kwargs['backtrace'] = RichTraceback()
            template_kwargs.update(kwargs)

            sickrage.app.log.error("%s: %s" % (str(kwargs['backtrace'].error.__class__.__name__), kwargs['backtrace'].error))

            return self.application.settings['templates']['errors/500.mako'].render_unicode(**template_kwargs)

    def render(self, template_name, **kwargs):
        self.write(self.render_string(template_name, **kwargs))

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')
        self.set_header('Cache-Control', 'max-age=0,no-cache,no-store')

    def redirect(self, url, permanent=True, status=None):
        if sickrage.app.config.web_root not in url:
            url = urljoin(sickrage.app.config.web_root + '/', url.lstrip('/'))
        super(BaseHandler, self).redirect(url, permanent, status)

    def previous_url(self):
        url = urlparse(self.request.headers.get("referer", "/{}/".format(sickrage.app.config.default_page)))
        return url._replace(scheme="", netloc="").geturl()

    def _genericMessage(self, subject, message):
        return self.render('generic_message.mako',
                           message=message,
                           subject=subject,
                           title="",
                           controller='root',
                           action='genericmessage')

    def get_url(self, url):
        if sickrage.app.config.web_root not in url:
            url = urljoin(sickrage.app.config.web_root + '/', url.lstrip('/'))
        url = urljoin("{}://{}".format(self.request.protocol, self.request.host), url)
        return url

    def run_in_executor(self, f, *args, **kwargs):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            threading.currentThread().setName('TORNADO')
            return f(*args, **kwargs)

        return sickrage.app.wserver.io_loop.run_in_executor(None, functools.partial(wrapper, *args, **kwargs))



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import json
from abc import ABC

from tornado.web import authenticated

import sickrage
from sickrage.core.helpers import get_internal_ip, get_external_ip
from sickrage.core.webserver.handlers.base import BaseHandler


class AccountLinkHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        code = self.get_argument('code', None)

        redirect_uri = "{}://{}{}/account/link".format(self.request.protocol, self.request.host, sickrage.app.config.web_root)

        if code:
            token = sickrage.app.auth_server.authorization_code(code, redirect_uri)
            if not token:
                return self.redirect('/account/link')

            certs = sickrage.app.auth_server.certs()
            if not certs:
                return self.redirect('/account/link')

            decoded_token = sickrage.app.auth_server.decode_token(token['access_token'], certs)
            if not decoded_token:
                return self.redirect('/account/link')

            if sickrage.app.api.token:
                sickrage.app.api.logout()

            sickrage.app.api.token = token

            sickrage.app.config.enable_sickrage_api = True

            if not sickrage.app.config.sub_id or not sickrage.app.config.server_id:
                sickrage.app.config.sub_id = decoded_token.get('sub')

                internal_connections = "{}://{}:{}{}".format(self.request.protocol,
                                                             get_internal_ip(),
                                                             sickrage.app.config.web_port,
                                                             sickrage.app.config.web_root)

                external_connections = "{}://{}:{}{}".format(self.request.protocol,
                                                             get_external_ip(),
                                                             sickrage.app.config.web_port,
                                                             sickrage.app.config.web_root)

                connections = ','.join([internal_connections, external_connections])

                server_id = sickrage.app.api.account.register_server(connections)
                if server_id:
                    sickrage.app.config.server_id = server_id

            sickrage.app.config.save()

            sickrage.app.alerts.message(_('Linked SiCKRAGE account to SiCKRAGE API'))
        else:
            authorization_url = sickrage.app.auth_server.authorization_url(redirect_uri=redirect_uri, scope="profile email offline_access")
            if authorization_url:
                return super(BaseHandler, self).redirect(authorization_url)

        return self.redirect('/account/link')


class AccountUnlinkHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        # if not sickrage.app.config.sub_id == self.get_current_user().get('sub'):
        #     return self.redirect("/{}/".format(sickrage.app.config.default_page))

        if not sickrage.app.config.server_id or sickrage.app.api.account.unregister_server(sickrage.app.config.server_id):
            sickrage.app.config.server_id = ""
            sickrage.app.config.sub_id = ""
            sickrage.app.api.logout()

            del sickrage.app.api.token

            sickrage.app.config.enable_sickrage_api = False
            sickrage.app.config.save()

            sickrage.app.alerts.message(_('Unlinked SiCKRAGE account from SiCKRAGE API'))


class AccountIsLinkedHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        return self.write(json.dumps({'linked': ('true', 'false')[not sickrage.app.api.userinfo]}))



import base64
import os
import zlib

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import load_pem_private_key, load_pem_public_key

import sickrage


def initialize():
    private_key_filename = os.path.join(sickrage.app.data_dir, 'privatekey.pem')
    sickrage.app.private_key = load_private_key(private_key_filename)
    if not sickrage.app.private_key:
        sickrage.app.private_key = generate_private_key()
        save_private_key(private_key_filename, sickrage.app.private_key)
    sickrage.app.public_key = sickrage.app.private_key.public_key()


def generate_private_key():
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend()
    )


def verify_public_key(public_key, private_key):
    try:
        decrypt_string(encrypt_string(b'sickrage', public_key), private_key) == b'sickrage'
    except ValueError:
        return False
    return True


def load_public_key(filename):
    if not os.path.exists(filename):
        return

    try:
        with open(filename, 'rb') as fd:
            public_key = load_pem_public_key(fd.read(), default_backend())
        return public_key
    except Exception:
        return


def load_private_key(filename):
    if not os.path.exists(filename):
        return

    try:
        with open(filename, 'rb') as fd:
            private_key = load_pem_private_key(fd.read(), None, default_backend())
        return private_key
    except Exception:
        return


def save_public_key(filename, public_key):
    pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )

    with open(filename, 'wb') as fd:
        fd.write(pem)


def save_private_key(filename, private_key):
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )

    with open(filename, 'wb') as fd:
        fd.write(pem)


def encrypt_file(filename, public_key):
    chunk_size = 245
    offset = 0
    end_loop = False
    encrypted = b""

    with open(filename, 'rb') as fd:
        blob = zlib.compress(fd.read())

    while not end_loop:
        chunk = blob[offset:offset + chunk_size]

        if len(chunk) % chunk_size != 0:
            end_loop = True
            chunk += b" " * (chunk_size - len(chunk))

        encrypted += public_key.encrypt(
            chunk,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        offset += chunk_size

    with open(filename, 'wb') as fd:
        fd.write(base64.b64encode(encrypted))


def decrypt_file(filename, private_key):
    chunk_size = 512
    offset = 0
    decrypted = b""

    with open(filename, 'rb') as fd:
        encrypted_blob = base64.b64decode(fd.read())

    while offset < len(encrypted_blob):
        chunk = encrypted_blob[offset: offset + chunk_size]

        decrypted += private_key.decrypt(
            chunk,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        offset += chunk_size

    with open(filename, 'wb') as fd:
        fd.write(zlib.decompress(decrypted))


def encrypt_string(string, public_key):
    chunk_size = 245
    offset = 0
    end_loop = False
    encrypted = b""

    blob = zlib.compress(string)
    while not end_loop:
        chunk = blob[offset:offset + chunk_size]

        if len(chunk) % chunk_size != 0:
            end_loop = True
            chunk += b" " * (chunk_size - len(chunk))

        encrypted += public_key.encrypt(
            chunk,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        offset += chunk_size

    return base64.b64encode(encrypted)


def decrypt_string(string, private_key):
    chunk_size = 512
    offset = 0
    decrypted = b""

    encrypted_blob = base64.b64decode(string)
    while offset < len(encrypted_blob):
        chunk = encrypted_blob[offset: offset + chunk_size]

        decrypted += private_key.decrypt(
            chunk,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        offset += chunk_size

    return zlib.decompress(decrypted)


# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import random
import string
import threading
from collections import deque
from enum import Enum
from functools import cmp_to_key

import sickrage


class TaskPriority(object):
    LOW = 10
    NORMAL = 20
    HIGH = 30
    EXTREME = 40


class TaskStatus(Enum):
    """
    Defines all the possible status for a task in the queue
    """
    QUEUED = 'queued'
    FINISHED = 'finished'
    FAILED = 'failed'
    STARTED = 'started'
    DEFERRED = 'deferred'
    NOT_QUEUED = 'not queued'


class WorkerStatus(Enum):
    """
    Defines all the possible status for a worker in the queue
    """
    WORKING = 'working'
    IDLE = 'idle'
    STOPPED = 'stopped'


class Queue(object):
    def __init__(self, name="QUEUE"):
        super(Queue, self).__init__()
        self.name = name
        self.lock = threading.RLock()
        self.queue = deque([])
        self.tasks = {}
        self.workers = []
        self.timer = None
        self.auto_remove_tasks_timer = None
        self.pause = False

    def start_worker(self, n_workers=1):
        """
        This function starts a given number of workers providing them a
        random identifier.
        :param n_workers: the total new workers to be created
        :return: the list of new workers
        """
        ids = []
        for i in range(0, n_workers):
            worker_id = "w" + self.get_random_id()
            self.workers.append(Worker(worker_id, self))
            ids.append(worker_id)

        self.auto_remove_tasks_timer = threading.Timer(10.0, self.auto_remove_tasks)
        self.auto_remove_tasks_timer.setName(self.name)
        self.auto_remove_tasks_timer.start()

        return ids

    def stop_worker(self, worker_id=None):
        """
        This function stops and kills a worker.
        If no id is provided, all workers are killed.
        :param worker_id: the identifier for the worker to kill
        :return: None
        """
        try:
            self.lock.acquire()
            if worker_id is None:
                sickrage.app.log.info("Shutting down all {} workers".format(self.name))
                for worker in self.workers:
                    worker.must_die = True
            else:
                for worker in self.workers:
                    if worker.id == worker_id:
                        worker.must_die = True
                        break

            if self.auto_remove_tasks_timer is not None:
                self.auto_remove_tasks_timer.cancel()
                self.auto_remove_tasks_timer = None
        finally:
            self.lock.release()
            self.notify_workers()

    def remove_worker(self, worker_id):
        """
        This function removes a worker from the list of workers
        (only if the worker was notified "to die" previously)
        :param worker_id: the ID for the worker to remove
        :return:
        """
        try:
            self.lock.acquire()

            i = 0

            for worker in self.workers:
                if worker.id == worker_id and worker.must_die is True:
                    self.workers.pop(i)
                    break
                i += 1
        finally:
            self.lock.release()
            self.notify_workers()

    def auto_remove_tasks(self):
        for task in self.tasks.copy().values():
            if task.status in [TaskStatus.FINISHED, TaskStatus.FAILED] and task.auto_remove:
                self.remove_task(task.id)

        self.auto_remove_tasks_timer = threading.Timer(10.0, self.auto_remove_tasks)
        self.auto_remove_tasks_timer.setName(self.name)
        self.auto_remove_tasks_timer.start()

    def get(self):
        def queue_sorter(x, y):
            """
            Sorts by priority descending then time ascending
            """
            if x.priority == y.priority:
                if y.added == x.added:
                    return 0
                elif y.added < x.added:
                    return 1
                elif y.added > x.added:
                    return -1
            else:
                return y.priority - x.priority

        try:
            self.lock.acquire()

            if len(self.queue) > 0:
                self.queue = deque(sorted(self.queue, key=cmp_to_key(lambda x, y: queue_sorter(x, y))))

                if self.is_paused:
                    if self.timer is None:
                        self.timer = threading.Timer(10.0, self.notify_workers)
                        self.timer.setName(self.name)
                        self.timer.start()
                    return None

                switch_pos = 1
                next_task = self.queue[len(self.queue) - 1]
                runnable = next_task.can_run(self.tasks)

                while not runnable:
                    switch_pos = switch_pos + 1
                    if switch_pos > len(self.queue):
                        self.queue.rotate(1)
                        switch_pos = 0
                        if self.timer is None:
                            self.timer = threading.Timer(10.0, self.notify_workers)
                            self.timer.setName(self.name)
                            self.timer.start()
                        return None
                    elif len(self.queue) > 1:
                        task_aux = self.queue[len(self.queue) - switch_pos]
                        self.queue[len(self.queue) - switch_pos] = self.queue[len(self.queue) - 1]
                        self.queue[len(self.queue) - 1] = task_aux
                    next_task = self.queue[len(self.queue) - 1]
                    runnable = next_task.can_run(self.tasks)

                return self.queue.pop()
            return None
        finally:
            self.lock.release()

    def put(self, task, task_id=None, depend=None, *args, **kwargs):
        """
        Adds an task to this queue

        :param task: Task object to add
        :param task_id: Task ID to add to task object
        :param depend: Task depends on other task to be in queue
        :return: task_id
        """
        try:
            self.lock.acquire()

            if not task_id:
                task_id = self.get_random_id()
                while task_id in self.tasks:
                    task_id = self.get_random_id()
            elif task_id in self.tasks:
                raise RuntimeError("Task already in {} (Task id : {})".format(self.name, task_id))

            task.id = task_id
            task.added = datetime.datetime.now()
            task.name = "{}-{}-{}".format(self.name, task_id, task.name)
            task.depend = depend

            self.tasks[task_id] = task
            self.queue.appendleft(task)

            sickrage.app.log.debug("New {} task {} added".format(self.name, task_id))
        finally:
            self.lock.release()
            self.notify_workers()
            return task_id

    def notify_workers(self):
        # sickrage.app.log.debug("Notifying {} workers".format(self.name))
        if self.timer is not None:
            # sickrage.app.log.debug("Clearing {} timer".format(self.name))
            self.timer.cancel()
            self.timer = None
        for worker in self.workers:
            worker.notify()

    def check_status(self, task_id):
        task = self.tasks.get(task_id, None)
        if task:
            return task.status
        return TaskStatus.NOT_QUEUED

    def fetch_task(self, task_id):
        return self.tasks.get(task_id, None)

    def remove_task(self, task_id):
        try:
            self.lock.acquire()

            if task_id in self.tasks:
                sickrage.app.log.debug("Removing {} task {}".format(self.name, task_id))
                task = self.tasks.get(task_id)
                if task in self.queue:
                    self.queue.remove(self.tasks.get(task_id))
                del self.tasks[task_id]
        finally:
            self.lock.release()

    def get_result(self, task_id, remove=True):
        task = self.tasks.get(task_id, None)
        if task:
            if remove and (task.status == TaskStatus.FINISHED or task.status == TaskStatus.FAILED):
                sickrage.app.log.debug("Removing {} task {}".format(self.name, task_id))
                self.tasks.pop(task_id)
            return task.result
        return TaskStatus.NOT_QUEUED

    @property
    def is_busy(self):
        return any(worker.status == WorkerStatus.WORKING for worker in self.workers)

    @property
    def is_paused(self):
        return self.pause

    def pause(self):
        """Pauses this queue"""
        sickrage.app.log.info("Pausing {}".format(self.name))
        self.pause = True

    def unpause(self):
        """Unpauses this queue"""
        sickrage.app.log.info("Un-pausing {}".format(self.name))
        self.pause = False

    def get_random_id(self):
        """
        This function returns a new random task id
        @returns taskID
        """
        return ''.join(random.sample(string.ascii_letters + string.octdigits * 5, 10)).upper()

    def shutdown(self):
        sickrage.app.log.info("Shutting down {}".format(self.name))
        self.stop_worker()
        self.queue.clear()
        self.tasks.clear()


class Worker(object):
    def __init__(self, _id, _queue):
        self.id = _id
        self.queue = _queue
        self.status = WorkerStatus.IDLE
        self.must_die = False
        self.task = None

    def notify(self):
        if self.status != WorkerStatus.WORKING:
            if self.must_die:
                self.queue.remove_worker(self.id)
            else:
                task = self.queue.get()
                if task is not None:
                    self.task = task
                    WorkerThread(self).start()

    def run(self):
        try:
            sickrage.app.log.debug("Worker " + str(self.id) + " task started...")
            self.status = WorkerStatus.WORKING
            # fn = self.task.fn
            # args = self.task.args
            self.task.status = TaskStatus.STARTED
            self.task.result = self.task.run()
            self.task.status = TaskStatus.FINISHED
        except Exception as e:
            if self.task is not None:
                self.task.status = TaskStatus.FAILED
                self.task.error_message = str(e)
            else:
                sickrage.app.log.debug("Worker " + str(self.id) + " without task.")
        finally:
            sickrage.app.log.debug("Worker " + str(self.id) + " task completed...")
            self.status = WorkerStatus.IDLE
            self.task = None
            self.notify()


class WorkerThread(threading.Thread):
    def __init__(self, worker):
        super(WorkerThread, self).__init__()
        self.name = worker.task.name
        self.worker = worker

    def run(self):
        self.worker.run()


class Task(object):
    def __init__(self, name, action_id=0, depend=None):
        super(Task, self).__init__()
        self.name = name.replace(" ", "-").upper()
        self.id = None
        self.action_id = action_id
        self.priority = TaskPriority.NORMAL
        self.status = TaskStatus.QUEUED
        self.added = None
        self.result = None
        self.error_message = None
        self.depend = depend
        self.auto_remove = True

    def is_finished(self):
        return self.status == TaskStatus.FINISHED

    def is_started(self):
        return self.status == TaskStatus.STARTED

    def is_failed(self):
        return self.status == TaskStatus.FAILED

    def get_status(self):
        return self.status

    def can_run(self, tasks):
        if self.depend is not None:
            for dependency in self.depend:
                task = tasks.get(dependency, None)
                if task is None:
                    # sickrage.app.log.debug("Cannot run task " + str(self.id) + ". Unable to find task " + str(dependency) + " in queue.")
                    return False
                if not task.is_finished():
                    # sickrage.app.log.debug("Cannot run task " + str(self.id) + ". Task " + str(dependency) + " is not finished")
                    return False

        return True



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import os
from abc import ABC

from tornado.web import authenticated

import sickrage
from sickrage.core.helpers import checkbox_to_value
from sickrage.core.nameparser import validator
from sickrage.core.webserver import ConfigHandler
from sickrage.core.webserver.handlers.base import BaseHandler


def is_naming_pattern_valid(pattern=None, multi=None, abd=None, sports=None, anime_type=None):
    if pattern is None:
        return 'invalid'

    if multi is not None:
        multi = int(multi)

    if anime_type is not None:
        anime_type = int(anime_type)

    # air by date shows just need one check, we don't need to worry about season folders
    if abd:
        is_valid = validator.check_valid_abd_naming(pattern)
        require_season_folders = False

    # sport shows just need one check, we don't need to worry about season folders
    elif sports:
        is_valid = validator.check_valid_sports_naming(pattern)
        require_season_folders = False

    else:
        # check validity of single and multi ep cases for the whole path
        is_valid = validator.check_valid_naming(pattern, multi, anime_type)

        # check validity of single and multi ep cases for only the file name
        require_season_folders = validator.check_force_season_folders(pattern, multi, anime_type)

    if is_valid and not require_season_folders:
        return 'valid'
    elif is_valid and require_season_folders:
        return 'seasonfolders'
    else:
        return 'invalid'


def is_rar_supported():
    """
    Test Packing Support:
        - Simulating in memory rar extraction on test.rar file
    """

    check = sickrage.app.config.change_unrar_tool(sickrage.app.config.unrar_tool, sickrage.app.config.unrar_alt_tool)

    if not check:
        sickrage.app.log.warning('Looks like unrar is not installed, check failed')

    return ('not supported', 'supported')[check]


class ConfigPostProcessingHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        return self.render('config/postprocessing.mako',
                           submenu=ConfigHandler.menu,
                           title=_('Config - Post Processing'),
                           header=_('Post Processing'),
                           topmenu='config',
                           controller='config',
                           action='postprocessing')


class SavePostProcessingHandler(BaseHandler, ABC):
    @authenticated
    async def post(self, *args, **kwargs):
        naming_pattern = self.get_argument('naming_pattern', '')
        naming_multi_ep = self.get_argument('naming_multi_ep', '')
        kodi_data = self.get_argument('kodi_data', '')
        kodi_12plus_data = self.get_argument('kodi_12plus_data', '')
        mediabrowser_data = self.get_argument('mediabrowser_data', '')
        sony_ps3_data = self.get_argument('sony_ps3_data', '')
        wdtv_data = self.get_argument('wdtv_data', '')
        tivo_data = self.get_argument('tivo_data', '')
        mede8er_data = self.get_argument('mede8er_data', '')
        keep_processed_dir = self.get_argument('keep_processed_dir', '')
        process_method = self.get_argument('process_method', '')
        del_rar_contents = self.get_argument('del_rar_contents', '')
        process_automatically = self.get_argument('process_automatically', '')
        no_delete = self.get_argument('no_delete', '')
        rename_episodes = self.get_argument('rename_episodes', '')
        airdate_episodes = self.get_argument('airdate_episodes', '')
        file_timestamp_timezone = self.get_argument('file_timestamp_timezone', '')
        unpack = self.get_argument('unpack', '')
        move_associated_files = self.get_argument('move_associated_files', '')
        sync_files = self.get_argument('sync_files', '')
        postpone_if_sync_files = self.get_argument('postpone_if_sync_files', '')
        nfo_rename = self.get_argument('nfo_rename', '')
        tv_download_dir = self.get_argument('tv_download_dir', '')
        naming_custom_abd = self.get_argument('naming_custom_abd', '')
        naming_anime = self.get_argument('naming_anime', '')
        create_missing_show_dirs = self.get_argument('create_missing_show_dirs', '')
        add_shows_wo_dir = self.get_argument('add_shows_wo_dir', '')
        naming_abd_pattern = self.get_argument('naming_abd_pattern', '')
        naming_strip_year = self.get_argument('naming_strip_year', '')
        delete_failed = self.get_argument('delete_failed', '')
        extra_scripts = self.get_argument('extra_scripts', '')
        naming_custom_sports = self.get_argument('naming_custom_sports', '')
        naming_sports_pattern = self.get_argument('naming_sports_pattern', '')
        naming_custom_anime = self.get_argument('naming_custom_anime', '')
        naming_anime_pattern = self.get_argument('naming_anime_pattern', '')
        naming_anime_multi_ep = self.get_argument('naming_anime_multi_ep', '')
        autopostprocessor_frequency = self.get_argument('autopostprocessor_frequency', '')
        delete_non_associated_files = self.get_argument('delete_non_associated_files', '')
        allowed_extensions = self.get_argument('allowed_extensions', '')
        processor_follow_symlinks = self.get_argument('processor_follow_symlinks', '')
        unpack_dir = self.get_argument('unpack_dir', '')

        results = []

        if not sickrage.app.config.change_tv_download_dir(tv_download_dir):
            results += [_("Unable to create directory ") + os.path.normpath(tv_download_dir) + _(", dir not changed.")]

        sickrage.app.config.change_autopostprocessor_freq(autopostprocessor_frequency)
        sickrage.app.config.process_automatically = checkbox_to_value(process_automatically)

        if unpack:
            if is_rar_supported() != "not supported":
                sickrage.app.config.unpack = checkbox_to_value(unpack)
                sickrage.app.config.unpack_dir = unpack_dir
            else:
                sickrage.app.config.unpack = 0
                results.append(_("Unpacking Not Supported, disabling unpack setting"))
        else:
            sickrage.app.config.unpack = checkbox_to_value(unpack)

        sickrage.app.config.no_delete = checkbox_to_value(no_delete)
        sickrage.app.config.keep_processed_dir = checkbox_to_value(keep_processed_dir)
        sickrage.app.config.create_missing_show_dirs = checkbox_to_value(create_missing_show_dirs)
        sickrage.app.config.add_shows_wo_dir = checkbox_to_value(add_shows_wo_dir)
        sickrage.app.config.process_method = process_method
        sickrage.app.config.delrarcontents = checkbox_to_value(del_rar_contents)
        sickrage.app.config.extra_scripts = [x.strip() for x in extra_scripts.split('|') if x.strip()]
        sickrage.app.config.rename_episodes = checkbox_to_value(rename_episodes)
        sickrage.app.config.airdate_episodes = checkbox_to_value(airdate_episodes)
        sickrage.app.config.file_timestamp_timezone = file_timestamp_timezone
        sickrage.app.config.move_associated_files = checkbox_to_value(move_associated_files)
        sickrage.app.config.sync_files = sync_files
        sickrage.app.config.postpone_if_sync_files = checkbox_to_value(postpone_if_sync_files)
        sickrage.app.config.allowed_extensions = ','.join({x.strip() for x in allowed_extensions.split(',') if x.strip()})
        sickrage.app.config.naming_custom_abd = checkbox_to_value(naming_custom_abd)
        sickrage.app.config.naming_custom_sports = checkbox_to_value(naming_custom_sports)
        sickrage.app.config.naming_custom_anime = checkbox_to_value(naming_custom_anime)
        sickrage.app.config.naming_strip_year = checkbox_to_value(naming_strip_year)
        sickrage.app.config.delete_failed = checkbox_to_value(delete_failed)
        sickrage.app.config.nfo_rename = checkbox_to_value(nfo_rename)
        sickrage.app.config.delete_non_associated_files = checkbox_to_value(delete_non_associated_files)
        sickrage.app.config.processor_follow_symlinks = checkbox_to_value(processor_follow_symlinks)

        if is_naming_pattern_valid(pattern=naming_pattern, multi=naming_multi_ep) != "invalid":
            sickrage.app.config.naming_pattern = naming_pattern
            sickrage.app.config.naming_multi_ep = int(naming_multi_ep)
            sickrage.app.config.naming_force_folders = validator.check_force_season_folders()
        else:
            results.append(_("You tried saving an invalid naming config, not saving your naming settings"))

        if is_naming_pattern_valid(pattern=naming_anime_pattern, multi=naming_anime_multi_ep, anime_type=naming_anime) != "invalid":
            sickrage.app.config.naming_anime_pattern = naming_anime_pattern
            sickrage.app.config.naming_anime_multi_ep = int(naming_anime_multi_ep)
            sickrage.app.config.naming_anime = int(naming_anime)
        else:
            results.append(_("You tried saving an invalid anime naming config, not saving your naming settings"))

        if is_naming_pattern_valid(pattern=naming_abd_pattern, abd=True) != "invalid":
            sickrage.app.config.naming_abd_pattern = naming_abd_pattern
        else:
            results.append(_("You tried saving an invalid air-by-date naming config, not saving your air-by-date settings"))

        if is_naming_pattern_valid(pattern=naming_sports_pattern, multi=naming_multi_ep, sports=True) != "invalid":
            sickrage.app.config.naming_sports_pattern = naming_sports_pattern
        else:
            results.append(_("You tried saving an invalid sports naming config, not saving your sports settings"))

        sickrage.app.metadata_providers['kodi'].config = kodi_data
        sickrage.app.metadata_providers['kodi_12plus'].config = kodi_12plus_data
        sickrage.app.metadata_providers['mediabrowser'].config = mediabrowser_data
        sickrage.app.metadata_providers['sony_ps3'].config = sony_ps3_data
        sickrage.app.metadata_providers['wdtv'].config = wdtv_data
        sickrage.app.metadata_providers['tivo'].config = tivo_data
        sickrage.app.metadata_providers['mede8er'].config = mede8er_data

        sickrage.app.config.save()

        if len(results) > 0:
            [sickrage.app.log.warning(x) for x in results]
            sickrage.app.alerts.error(_('Error(s) Saving Configuration'), '<br>\n'.join(results))
        else:
            sickrage.app.alerts.message(_('[POST-PROCESSING] Configuration Encrypted and Saved to disk'))

        return self.redirect("/config/postProcessing/")


class TestNamingHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        pattern = self.get_argument('pattern', None)
        multi = self.get_argument('multi', None)
        abd = self.get_argument('abd', None)
        sports = self.get_argument('sports', None)
        anime_type = self.get_argument('anime_type', None)

        if multi is not None:
            multi = int(multi)

        if anime_type is not None:
            anime_type = int(anime_type)

        result = validator.test_name(pattern, multi, abd, sports, anime_type)

        result = os.path.join(result['dir'], result['name'])

        return self.write(result)


class IsNamingPatternValidHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        pattern = self.get_argument('pattern', None)
        multi = self.get_argument('multi', None)
        abd = self.get_argument('abd', None)
        sports = self.get_argument('sports', None)
        anime_type = self.get_argument('anime_type', None)

        return self.write(is_naming_pattern_valid(pattern=pattern, multi=multi, abd=abd, sports=sports, anime_type=anime_type))


class IsRarSupportedHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        return self.write(is_rar_supported())



# coding=utf-8
# Author: Gonçalo M. (aka duramato/supergonkas) <supergonkas@gmail.com>
#
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE. If not, see <http://www.gnu.org/licenses/>.


import re

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, try_int, convert_size
from sickrage.providers import TorrentProvider


class LimeTorrentsProvider(TorrentProvider):
    def __init__(self):
        super(LimeTorrentsProvider, self).__init__('LimeTorrents', 'https://www.limetorrents.cc', False)

        self._urls.update({
            'update': '{base_url}/post/updatestats.php'.format(**self._urls),
            'search': '{base_url}/search/tv/%s/'.format(**self._urls),
            'rss': '{base_url}/browse-torrents/TV-shows/'.format(**self._urls),
        })

        self.minseed = None
        self.minleech = None
        self.confirmed = False

        self.proper_strings = ['PROPER', 'REPACK', 'REAL']

        self.cache = TVCache(self)

    def search(self, search_strings, age=0, show_id=None, season=None, episode=None, **kwargs):
        results = []
        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {}".format(mode))
            for search_string in search_strings[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: {}".format(search_string))

                search_url = (self.urls['rss'], self.urls['search'] % search_string)[mode != 'RSS']

                resp = self.session.get(search_url)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        id_regex = re.compile(r'(?:\/)(.*)(?:-torrent-([0-9]*)\.html)', re.I)
        hash_regex = re.compile(r'(.*)([0-9a-f]{40})(.*)', re.I)

        def process_column_header(th):
            return th.span.get_text() if th.span else th.get_text()

        with bs4_parser(data) as html:
            torrent_table = html.find('table', class_='table2')

            if not torrent_table:
                sickrage.app.log.debug('Data returned from provider does not contain any torrents')
                return results

            torrent_rows = torrent_table.find_all('tr')
            labels = [process_column_header(label) for label in torrent_rows[0].find_all('th')]

            # Skip the first row, since it isn't a valid result
            for row in torrent_rows[1:]:
                cells = row.find_all('td')

                try:
                    title_cell = cells[labels.index('Torrent Name')]

                    verified = title_cell.find('img', title='Verified torrent')
                    if self.confirmed and not verified:
                        continue

                    title_anchors = title_cell.find_all('a')
                    if not title_anchors or len(title_anchors) < 2:
                        continue

                    title_url = title_anchors[0].get('href')
                    title = title_anchors[1].get_text(strip=True)
                    regex_result = id_regex.search(title_anchors[1].get('href'))

                    alt_title = regex_result.group(1)
                    if len(title) < len(alt_title):
                        title = alt_title.replace('-', ' ')

                    torrent_id = regex_result.group(2)
                    info_hash = hash_regex.search(title_url).group(2)
                    if not all([title, torrent_id, info_hash]):
                        continue

                    try:
                        self.session.get(self.urls['update'], timeout=30,
                                         params={'torrent_id': torrent_id, 'infohash': info_hash})
                    except Exception:
                        pass

                    download_url = 'magnet:?xt=urn:btih:{hash}&dn={title}'.format(hash=info_hash, title=title)

                    # Remove comma as thousands separator from larger number like 2,000 seeders = 2000
                    seeders = try_int(cells[labels.index('Seed')].get_text(strip=True).replace(',', ''), 1)
                    leechers = try_int(cells[labels.index('Leech')].get_text(strip=True).replace(',', ''))

                    size = convert_size(cells[labels.index('Size')].get_text(strip=True), -1)

                    results += [
                        {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                    ]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider")

        return results
